use microrm::{
    prelude::{Insertable, Queryable},
    schema::Serializable,
};
use serde::Deserialize;

use crate::{
    config::Config,
    schema::{self, UIDCDatabase},
    server::{ServerStateWrapper, SessionHelper, UIDCRequest},
};

#[derive(Debug, Clone, Deserialize)]
#[serde(deny_unknown_fields)]
pub struct GithubConfig {
    pub login_url: Option<String>,
    pub token_url: Option<String>,
    pub api_base: Option<String>,

    pub client_id: String,
    pub client_secret: String,
}

const DEFAULT_LOGIN_URL: &'static str = "https://github.com/login/oauth/authorize";
const DEFAULT_TOKEN_URL: &'static str = "https://github.com/login/oauth/access_token";
const DEFAULT_API_BASE: &'static str = "https://api.github.com";

#[derive(Deserialize)]
#[serde(rename_all = "lowercase")]
pub enum CallbackRequestType {
    Login,
    Register,
}

pub struct GithubAuthenticator {
    base_url: String,
    config: GithubConfig,
}

impl super::ExternalAuthenticator for GithubAuthenticator {
    fn build(_db: &UIDCDatabase, config: &Config) -> Option<Self> {
        config.github.as_ref().map(|ghc| Self {
            base_url: config.base_url.clone(),
            config: ghc.clone(),
        })
    }

    fn register_routes(&'static self, server: &mut tide::Server<ServerStateWrapper>) {
        server
            .at("/:realm/github_return")
            .get(|req: UIDCRequest| async { Ok(self.extract_login_state(req).await) });
    }

    fn generate_login_url(&self, realm: &str, redirect: &str) -> String {
        let return_url = format!(
            "{base_url}/{realm}/github_return?redirect={redirect}&mode=login",
            base_url = self.base_url,
        );

        tide::http::Url::parse_with_params(
            self.config
                .login_url
                .as_deref()
                .unwrap_or(DEFAULT_LOGIN_URL),
            &[
                ("client_id", self.config.client_id.as_str()),
                ("redirect_uri", return_url.as_str()),
            ],
        )
        .expect("couldn't generate login url for github")
        .into()
    }

    fn generate_registration_url(&self, realm: &str, redirect: &str) -> String {
        let return_url = format!(
            "{base_url}/{realm}/github_return?redirect={redirect}&mode=register",
            base_url = self.base_url,
        );

        tide::http::Url::parse_with_params(
            self.config
                .login_url
                .as_deref()
                .unwrap_or(DEFAULT_LOGIN_URL),
            &[
                ("client_id", self.config.client_id.as_str()),
                ("redirect_uri", return_url.as_str()),
            ],
        )
        .expect("couldn't generate login url for github")
        .into()
    }

    fn extract_login_state(
        &self,
        req: UIDCRequest,
    ) -> impl smol::prelude::Future<Output = tide::Response> {
        async move {
            let state = req.state();
            let realm = req.param("realm").unwrap();

            #[derive(Deserialize)]
            struct Query {
                code: String,
                redirect: String,
                mode: CallbackRequestType,
            }
            let Ok(query) = req.query::<Query>() else {
                return tide::Response::builder(400)
                    .body("Query string invalid.")
                    .build();
            };

            #[derive(Deserialize)]
            struct TokenResponse {
                access_token: String,
            }

            let auth = surf::http::auth::BasicAuth::new(
                self.config.client_id.as_str(),
                self.config.client_secret.as_str(),
            );

            let resp: TokenResponse = match state
                .client
                .post(
                    tide::http::Url::parse_with_params(
                        self.config
                            .token_url
                            .as_deref()
                            .unwrap_or(DEFAULT_TOKEN_URL),
                        &[
                            ("client_id", self.config.client_id.as_str()),
                            ("client_secret", self.config.client_secret.as_str()),
                            ("code", query.code.as_str()),
                        ],
                    )
                    .expect("couldn't generate token url for github"),
                )
                .header(auth.name(), auth.value())
                .content_type(surf::http::mime::FORM)
                .recv_form()
                .await
            {
                Ok(resp) => resp,
                Err(err) => {
                    return tide::Response::builder(500)
                        .body(format!("could not parse Github response for token: {err}"))
                        .build()
                }
            };

            let atoken = resp.access_token;

            #[derive(Deserialize)]
            struct UserInfoResponse {
                id: i64,
            }

            let resp: UserInfoResponse = match state
                .client
                .get(format!(
                    "{base}/user",
                    base = self.config.api_base.as_deref().unwrap_or(DEFAULT_API_BASE)
                ))
                .header("Authorization", format!("Bearer {atoken}"))
                .content_type(surf::http::mime::JSON)
                .recv_json()
                .await
            {
                Ok(resp) => resp,
                Err(err) => {
                    return tide::Response::builder(500)
                        .body(format!("could not parse Github response for token: {err}"))
                        .build()
                }
            };

            let user_id = resp.id.to_string();

            let Some(realm) = state.db.realms.keyed(realm).get().ok().flatten() else {
                return tide::Response::builder(404).body("no such realm").build();
            };

            let external_auth_map = realm
                .external_auth
                .keyed((
                    &user_id,
                    schema::ExternalAuthProvider::Github.into_serialized(),
                ))
                .get()
                .ok()
                .flatten();

            match (query.mode, external_auth_map) {
                (CallbackRequestType::Login, Some(map)) => {
                    return self.handle_matching_login(
                        req,
                        map.internal_user_id,
                        query.redirect.as_str(),
                    );
                }
                (CallbackRequestType::Login, None) => {
                    return self.handle_no_mapping(req, query.redirect);
                }
                (CallbackRequestType::Register, _) => {
                    todo!()
                }
            }
        }
    }
}
